/*
    server.c - OSS trampoline sample server

    Copyright (C) 2000 Ram Sivrasubramanium <ram@linux-workshop.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "dsproxy.h"
#include <sys/time.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>


/*
 * read exactly "count" number of bytes from a file descriptor
 */
static int
read_exact (int fd, void *buf2, size_t count)
{
  char *buf = buf2;
  int savedcount = count;
  size_t siz;

  while (count > 0) {
    siz = read (fd, buf, count);
    if (siz < 0)
      return (siz);

    buf += siz;
    count -= siz;
  }

  return (savedcount);
}


/*
 * establish network connection
 */
static int
ListenOnTCPPort (int port)
{
  struct sockaddr_in addr;
  int sock;
  int one = 1;

  addr.sin_family = AF_INET;
  addr.sin_port = htons (port);
  addr.sin_addr.s_addr = htonl (INADDR_ANY);
  if ((sock = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    return (-1);
  if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof (one))
      < 0)
    return (-1);
  if (bind (sock, (struct sockaddr *) &addr, sizeof (addr)) < 0)
    return (-1);
  if (listen (sock, 5) < 0)
    return (-1);

  return (sock);
}


static int
AcceptConnectionOnTCPPort (int listensock)
{
  int sock;
#if !defined( __socklen_t_defined )
  struct sockaddr_in addr;
  int addrlen = sizeof (addr);
#else
  struct sockaddr addr;
  socklen_t addrlen = sizeof (addr);
#endif

  if ((sock = accept (listensock, &addr, &addrlen)) < 0)
    return (-1);


  if (fork () != 0)
    {
      close (sock);
      return 0;
    }

  return (sock);
}

int exit_all=0;

/*
 * simply copy bytes from dsproxy's /dev/dsp and reroute them across the network
 */
int
main (int argc, char **argv)
{
  int listensock, sock, fdi;
  char op;
  unsigned short len;
  uid_t uid;
  int cmd, val;
  char buffer[DSPROXY_FIFO_SIZE];
  fd_set set;
  struct timeval timeout;
  
  if ((listensock = ListenOnTCPPort (DSPROXY_TCP_PORT)) < 0)
    exit (-1);

  signal (SIGCHLD, SIG_IGN);



  for (;;)
    {
      fprintf (stderr, "accepting new client on port %i...\n",
	       DSPROXY_TCP_PORT);
      if ((sock = AcceptConnectionOnTCPPort (listensock)) < 0)
	{
	  fprintf (stderr, "Socket error, exiting [%i].\n", getpid ());
	  exit (255);
	}

      if (sock != 0) {	/* child */
	  if (!read (sock, &uid, sizeof (uid_t)))
	    {
	      fprintf (stderr, "couldn't read client UID\n");
	      exit (255);
	    }

	  fprintf (stderr,
		   "Client (UID %i) connected to server [PID %i] on socket %i.\n",
		   uid, getpid (), sock);

	  setuid (uid);		/* drop root rights */

	  if ((fdi = open ("/dev/dsp", O_RDONLY|O_NONBLOCK)) < 0)
	    {
	      fprintf (stderr, "Error in opening /dev/dsp!\n");
	      exit (255);
	    }

	  while (!exit_all)
	    {
	      /* Initialize the file descriptor set. */
	      FD_ZERO (&set);
	      FD_SET (fdi, &set);
	      FD_SET (sock, &set);

	      /* Initialize the timeout data structure. */
	      timeout.tv_sec = 1;
	      timeout.tv_usec = 0;
	      /* select returns 0 if timeout, 1 if input available, -1 if error. */
	      if (select (FD_SETSIZE,&set, NULL, NULL,&timeout)) {

		if (FD_ISSET(sock,&set)) {
		  read (sock, &op, sizeof (char));
		  switch (op) {
		  case DSPROXY_READER_DISCONNECT:
		    fprintf(stderr,"D");
		    exit_all = 1;
		    break;
		  }
		}


		if (FD_ISSET(fdi,&set)) {   
		  read_exact (fdi, &op, sizeof (char));
		
		  write (sock, &op, sizeof (char));
		  switch (op)
		    {
		    case DSPROXY_STREAM_DSPDATA:
		      fprintf (stderr,".");
		      read_exact (fdi, &len, sizeof (short));
		      read_exact (fdi, buffer, (len == 0) ? 65536 : len);
		      
		      write (sock, &len, sizeof (short));
		      write (sock, buffer, (len == 0) ? 65536 : len);
		      /* this is ok since /dev/dsp tags with exactly how many bytes follow */
		      break;

		    case DSPROXY_STREAM_IOCTL:
		    case DSPROXY_MIXER_IOCTL:
		      fprintf (stderr,"x");
		      /*
		      switch(cmd) {
		      case SNDCTL_DSP_RESET:
			fprintf(stderr, "SNDCTL_DSP_RESET\n");
			break;
			
		      case SNDCTL_DSP_POST:
			fprintf(stderr, "SNDCTL_DSP_POST\n");
			break;
	
		      case SNDCTL_DSP_SPEED:
			fprintf(stderr, "SNDCTL_DSP_SPEED    => %d\n", val);
						break;
	
		      case SNDCTL_DSP_STEREO:
			fprintf(stderr, "SNDCTL_DSP_STEREO   => %s\n", val?"yes":"no");
			break;
			
		      case SNDCTL_DSP_CHANNELS:
			fprintf(stderr, "SNDCTL_DSP_CHANNELS => %d\n", val);
			break;
	
		      case SNDCTL_DSP_SETFRAGMENT:
			fprintf(stderr, "SNDCTL_DSP_SETFRAGMENT => %08x\n", val);
			break;
			
		      case SNDCTL_DSP_SETFMT:
			fprintf(stderr, "SNDCTL_DSP_SETFMT   =>");
			if(val&AFMT_U8) fprintf(stderr, " AFMT_U8");
			if(val&AFMT_S8) fprintf(stderr, " AFMT_S8");
			if(val&AFMT_S16_LE) fprintf(stderr, " AFMT_S16_LE");
			if(val&AFMT_S16_BE) fprintf(stderr, " AFMT_S16_BE");
			if(val&AFMT_S16_NE) fprintf(stderr, " AFMT_S16_NE");
			if(val&AFMT_U16_LE) fprintf(stderr, " AFMT_U16_LE");
			if(val&AFMT_U16_BE) fprintf(stderr, " AFMT_U16_BE");
			fprintf(stderr, "\n");
			break;
		      case SOUND_MIXER_WRITE_VOLUME:
			fprintf(stderr, "SOUND_MIXER_WRITE_VOLUME => %d - %d\n", (val>>8)&0xff, val&0xff);
			break;

		      case SOUND_MIXER_WRITE_BASS:
			fprintf(stderr, "SOUND_MIXER_WRITE_BASS   => %d - %d\n", (val>>8)&0xff, val&0xff);
			break;

		      case SOUND_MIXER_WRITE_TREBLE:
			fprintf(stderr, "SOUND_MIXER_WRITE_TREBLE => %d - %d\n", (val>>8)&0xff, val&0xff);
			break;

		      case SOUND_MIXER_WRITE_PCM:
			fprintf(stderr, "SOUND_MIXER_WRITE_PCM    => %d - %d\n", (val>>8)&0xff, val&0xff);
			break;
			
		      case SOUND_MIXER_WRITE_OGAIN:
			fprintf(stderr, "SOUND_MIXER_WRITE_PCM    => %d - %d\n", (val>>8)&0xff, val&0xff);
			break;
		      }
		      */

		      read_exact (fdi, &cmd, sizeof (int));
		      read_exact (fdi, &val, sizeof (int));

		      write (sock, &cmd, sizeof (int));
		      write (sock, &val, sizeof (int));
		      break;

		    default:
		      break;
		    }
		}
	      }
	    }
	  fprintf(stderr,"server [%i] exiting\n",getpid());
	  close(fdi);
	  close(sock);
	  exit (0);		/* spawned child server proc did his task */
	}
    }
}
